"""

 torch_test.py

 Simple script to verify CUDA availability, used to verify a CUDA/PyTorch
 environment.

"""

def torch_test():
    """
    Print diagnostic information about Torch/CUDA status, including Torch/CUDA versions
    and all available CUDA device names.
    """
    
    import torch

    print('Torch version: {}'.format(str(torch.__version__)))
    print('CUDA available: {}'.format(torch.cuda.is_available()))

    device_ids = list(range(torch.cuda.device_count()))
    print('Found {} CUDA devices:'.format(len(device_ids)))
    for device_id in device_ids:
        device_name = 'unknown'
        try:
            device_name = torch.cuda.get_device_name(device=device_id)
        except Exception as e:
            pass
        print('{}: {}'.format(device_id,device_name))
       
if __name__ == '__main__':    
    torch_test()



"""

visualize_detector_output.py

Render images with bounding boxes annotated on them to a folder, based on a
detector output result file (.json), optionally writing an HTML index file.

"""

#%% Imports

import argparse
import json
import os
import random
import sys

from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from functools import partial
from tqdm import tqdm

from megadetector.data_management.annotations.annotation_constants import detector_bbox_category_id_to_name
from megadetector.visualization import visualization_utils as vis_utils
from megadetector.utils.ct_utils import get_max_conf
from megadetector.utils import write_html_image_list
from megadetector.detection.run_detector import get_typical_confidence_threshold_from_results


#%% Constants

# This will only be used if a category mapping is not available in the results file.
DEFAULT_DETECTOR_LABEL_MAP = {
    str(k): v for k, v in detector_bbox_category_id_to_name.items()
}


#%% Support functions

def _render_image(entry,
                 detector_label_map,classification_label_map,
                 confidence_threshold,classification_confidence_threshold,
                 render_detections_only,preserve_path_structure,out_dir,images_dir,
                 output_image_width):
    """
    Internal function for rendering a single image.
    """
    
    rendering_result = {'failed_image':False,'missing_image':False,
                        'skipped_image':False,'annotated_image_path':None,
                        'max_conf':None,'file':entry['file']}
    
    image_id = entry['file']

    if 'failure' in entry and entry['failure'] is not None:
        rendering_result['failed_image'] = True
        return rendering_result

    assert 'detections' in entry and entry['detections'] is not None
    
    max_conf = get_max_conf(entry)
    rendering_result['max_conf'] = max_conf
    
    if (max_conf < confidence_threshold) and render_detections_only:
        rendering_result['skipped_image'] = True
        return rendering_result
    
    image_obj = os.path.join(images_dir, image_id)
    if not os.path.exists(image_obj):
        print(f'Image {image_id} not found in images_dir')
        rendering_result['missing_image'] = True
        return rendering_result

    # If output_image_width is -1 or None, this will just return the original image
    image = vis_utils.resize_image(
        vis_utils.open_image(image_obj), output_image_width)

    vis_utils.render_detection_bounding_boxes(
        entry['detections'], image, 
        label_map=detector_label_map,
        classification_label_map=classification_label_map,
        confidence_threshold=confidence_threshold,
        classification_confidence_threshold=classification_confidence_threshold)

    if not preserve_path_structure:
        for char in ['/', '\\', ':']:
            image_id = image_id.replace(char, '~')
        annotated_img_path = os.path.join(out_dir, f'anno_{image_id}')
    else:
        assert not os.path.isabs(image_id), "Can't preserve paths when operating on absolute paths"
        annotated_img_path = os.path.join(out_dir, image_id)
        os.makedirs(os.path.dirname(annotated_img_path),exist_ok=True)
        
    image.save(annotated_img_path)
    rendering_result['annotated_image_path'] = annotated_img_path        

    return rendering_result


#%% Main function

def visualize_detector_output(detector_output_path,
                              out_dir,
                              images_dir,
                              confidence_threshold=0.15,
                              sample=-1,
                              output_image_width=700,
                              random_seed=None,
                              render_detections_only=False,
                              classification_confidence_threshold=0.1,
                              html_output_file=None,
                              html_output_options=None,
                              preserve_path_structure=False,
                              parallelize_rendering=False,
                              parallelize_rendering_n_cores=10,
                              parallelize_rendering_with_threads=True):
    
    """
    Draws bounding boxes on images given the output of a detector.

    Args:
        detector_output_path (str): path to detector output .json file
        out_dir (str): path to directory for saving annotated images
        images_dir (str): folder where the images live; filenames in 
            [detector_output_path] should be relative to [image_dir]
        confidence_threshold (float, optional): threshold above which detections will be rendered
        sample (int, optional): maximum number of images to render, -1 for all
        output_image_width (int, optional): width in pixels to resize images for display,
            preserving aspect ration; set to -1 to use original image width
        random_seed (int, optional): seed to use for choosing images when sample != -1
        render_detections_only (bool): only render images with above-threshold detections
        classification_confidence_threshold (float, optional): only show classifications
            above this threshold; does not impact whether images are rendered, only whether
            classification labels (not detection categories) are displayed
        html_output_file (str, optional): output path for an HTML index file (not written
            if None)
        html_output_options (dict, optional): HTML formatting options; see write_html_image_list 
            for details
        preserve_path_structure (bool, optional): if False (default), writes images to unique
            names in a flat structure in the output folder; if True, preserves relative paths
            within the output folder
        parallelize_rendering (bool, optional): whether to use concurrent workers for rendering
        parallelize_rendering_n_cores (int, optional): number of concurrent workers to use 
            (ignored if parallelize_rendering is False)
        parallelize_rendering_with_threads (bool, optional): determines whether we use
            threads (True) or processes (False) for parallelization (ignored if parallelize_rendering 
            is False)

    Returns:
        list: list of paths to annotated images
    """
    
    assert os.path.exists(detector_output_path), \
        'Detector output file does not exist at {}'.format(detector_output_path)

    assert os.path.isdir(images_dir), \
        'Image folder {} is not available'.format(images_dir)

    os.makedirs(out_dir, exist_ok=True)


    ##%% Load detector output

    with open(detector_output_path) as f:
        detector_output = json.load(f)
    assert 'images' in detector_output, (
        'Detector output file should be a json with an "images" field.')
    images = detector_output['images']
    
    if confidence_threshold is None:
        confidence_threshold = get_typical_confidence_threshold_from_results(detector_output)
        
    assert confidence_threshold >= 0 and confidence_threshold <= 1, (
        f'Confidence threshold {confidence_threshold} is invalid, must be in (0, 1).')
    
    if 'detection_categories' in detector_output:
        detector_label_map = detector_output['detection_categories']
    else:
        detector_label_map = DEFAULT_DETECTOR_LABEL_MAP

    num_images = len(images)
    print(f'Detector output file contains {num_images} entries.')

    if sample > 0:
        assert num_images >= sample, (
            f'Sample size {sample} greater than number of entries '
            f'({num_images}) in detector result.')

        if random_seed is not None:
            images = sorted(images, key=lambda x: x['file'])
            random.seed(random_seed)

        random.shuffle(images)
        images = sorted(images[:sample], key=lambda x: x['file'])
        print(f'Sampled {len(images)} entries from the detector output file.')


    ##%% Load images, annotate them and save

    print('Rendering detections above a confidence threshold of {}'.format(
        confidence_threshold))
    
    classification_label_map = None
    
    if 'classification_categories' in detector_output:
        classification_label_map = detector_output['classification_categories']
        
    rendering_results = []
    
    if parallelize_rendering:
        
        if parallelize_rendering_with_threads:
            worker_string = 'threads'
        else:
            worker_string = 'processes'
            
        if parallelize_rendering_n_cores is None:
            if parallelize_rendering_with_threads:
                pool = ThreadPool()
            else:
                pool = Pool()
        else:
            if parallelize_rendering_with_threads:
                pool = ThreadPool(parallelize_rendering_n_cores)
            else:
                pool = Pool(parallelize_rendering_n_cores)
            print('Rendering images with {} {}'.format(parallelize_rendering_n_cores,
                                                       worker_string))            
        rendering_results = list(tqdm(pool.imap(
                                 partial(_render_image,detector_label_map=detector_label_map,
                                         classification_label_map=classification_label_map,
                                         confidence_threshold=confidence_threshold,
                                         classification_confidence_threshold=classification_confidence_threshold,
                                         render_detections_only=render_detections_only,
                                         preserve_path_structure=preserve_path_structure,
                                         out_dir=out_dir,
                                         images_dir=images_dir,
                                         output_image_width=output_image_width),
                                 images), total=len(images)))
        
    else:
                
        for entry in tqdm(images):
            
            rendering_result = _render_image(entry,detector_label_map,classification_label_map,
                                            confidence_threshold,classification_confidence_threshold,
                                            render_detections_only,preserve_path_structure,out_dir,
                                            images_dir,output_image_width)
            rendering_results.append(rendering_result)
        
    # ...for each image
    
    failed_images = [r for r in rendering_results if r['failed_image']]
    missing_images = [r for r in rendering_results if r['missing_image']]
    skipped_images = [r for r in rendering_results if r['skipped_image']]
    
    print('Skipped {} failed images (of {})'.format(len(failed_images),len(images)))
    print('Skipped {} missing images (of {})'.format(len(missing_images),len(images)))
    print('Skipped {} below-threshold images (of {})'.format(len(skipped_images),len(images)))
    
    print(f'Rendered detection results to {out_dir}')

    annotated_image_paths = [r['annotated_image_path'] for r in rendering_results if \
                             r['annotated_image_path'] is not None]
    
    if html_output_file is not None:
        
        html_dir = os.path.dirname(html_output_file)
        
        html_image_info = []
        
        for r in rendering_results:
            d = {}
            annotated_image_path_relative = os.path.relpath(r['annotated_image_path'],html_dir)
            d['filename'] = annotated_image_path_relative
            d['textStyle'] = \
             'font-family:verdana,arial,calibri;font-size:80%;' + \
                 'text-align:left;margin-top:20;margin-bottom:5'
            d['title'] = '{} (max conf: {})'.format(r['file'],r['max_conf'])
            html_image_info.append(d)
            
        _ = write_html_image_list.write_html_image_list(html_output_file,html_image_info,
                                                    options=html_output_options)
        
    return annotated_image_paths


#%% Command-line driver

def main():
    
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Annotate the bounding boxes predicted by a detector above '
                    'some confidence threshold, and save the annotated images.')
    parser.add_argument(
        'detector_output_path', type=str,
        help='Path to json output file of the detector')
    parser.add_argument(
        'out_dir', type=str,
        help='Path to directory where the annotated images will be saved. '
             'The directory will be created if it does not exist.')
    parser.add_argument(
        '-c', '--confidence', type=float, default=0.15,
        help='Value between 0 and 1, indicating the confidence threshold '
             'above which to visualize bounding boxes')
    parser.add_argument(
        '-i', '--images_dir', type=str, default=None,
        help='Path to a local directory where images are stored. This '
             'serves as the root directory for image paths in '
             'detector_output_path.')
    parser.add_argument(
        '-n', '--sample', type=int, default=-1,
        help='Number of images to be annotated and rendered. Set to -1 '
             '(default) to annotate all images in the detector output file. '
             'There may be fewer images if some are not found in images_dir.')
    parser.add_argument(
        '-w', '--output_image_width', type=int, default=700,
        help='Integer, desired width in pixels of the output annotated images. '
             'Use -1 to not resize. Default: 700.')
    parser.add_argument(
        '-r', '--random_seed', type=int, default=None,
        help='Integer, for deterministic order of image sampling')
    parser.add_argument(
        '-html', '--html_output_file', type=str, default=None,
        help='Filename to which we should write an HTML image index (off by default)')
    parser.add_argument(
        '--open_html_output_file', action='store_true',
        help='Open the .html output file when done')
    parser.add_argument(
        '-do', '--detections_only', action='store_true',
        help='Only render images with above-threshold detections (by default, '
             'both empty and non-empty images are rendered).')
    parser.add_argument(
        '-pps', '--preserve_path_structure', action='store_true',
        help='Preserve relative image paths (otherwise flattens and assigns unique file names)')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    visualize_detector_output(
        detector_output_path=args.detector_output_path,
        out_dir=args.out_dir,
        confidence_threshold=args.confidence,
        images_dir=args.images_dir,
        sample=args.sample,
        output_image_width=args.output_image_width,
        random_seed=args.random_seed,
        render_detections_only=args.detections_only,
        preserve_path_structure=args.preserve_path_structure,
        html_output_file=args.html_output_file)

    if args.html_output_file is not None and args.open_html_output_file:
        from megadetector.utils.path_utils import open_file
        open_file(args.html_output_file)

if __name__ == '__main__':
    main()


#%% Interactive driver

if False:
    
    pass

    #%%
    
    detector_output_path = os.path.expanduser('~/postprocessing/bellevue-camera-traps/bellevue-camera-traps-2023-12-05-v5a.0.0/combined_api_outputs/bellevue-camera-traps-2023-12-05-v5a.0.0_detections.json')
    out_dir = r'g:\temp\preview'
    images_dir = r'g:\camera_traps\camera_trap_images'
    confidence_threshold = 0.15
    sample = 50
    output_image_width = 700
    random_seed = 1
    render_detections_only = True
    classification_confidence_threshold = 0.1
    html_output_file = os.path.join(out_dir,'index.html')
    html_output_options = None
    preserve_path_structure = False
    parallelize_rendering = True
    parallelize_rendering_n_cores = 10
    parallelize_rendering_with_threads = False

    _ = visualize_detector_output(detector_output_path,
                              out_dir,
                              images_dir,
                              confidence_threshold,
                              sample,
                              output_image_width,
                              random_seed,
                              render_detections_only,
                              classification_confidence_threshold,
                              html_output_file,
                              html_output_options,
                              preserve_path_structure,
                              parallelize_rendering,
                              parallelize_rendering_n_cores,
                              parallelize_rendering_with_threads)
    
    from megadetector.utils.path_utils import open_file
    open_file(html_output_file)



"""

import_desert_lion_conservation_camera_traps.py

Prepare the Desert Lion Conservation Camera Traps dataset for release on LILA.

"""

#%% Imports and constants

import os
import json

input_base_folder = r'i:/data/desert-lion'
assert os.path.isdir(input_base_folder)

# md_results_file = r'i:/data/desert-lion/desert-lion-camera-traps-2024-07-14-v5a.0.0_detections-all.json'
md_results_file = r'i:/data/desert-lion/desert-lion-camera-traps-2024-07-14-v5a.0.0_detections.json'
assert os.path.isfile(md_results_file)

export_base = os.path.join(input_base_folder,'annotated-imgs')
assert os.path.isdir(export_base)

preview_dir = r'g:\temp\desert-lion-viz'
output_file = os.path.join(input_base_folder,'desert_lion_camera_traps.json')
output_zipfile = os.path.join(input_base_folder,'desert-lion-camera-traps-images.zip')

exif_cache_file_post_exif_removal = os.path.join(input_base_folder,'exif_data_post_exif_removal.json')
exif_cache_file = os.path.join(input_base_folder,'exif_data.json')


#%% Find images and videos

from megadetector.detection.video_utils import find_videos
from megadetector.utils.path_utils import find_images

video_files = find_videos(input_base_folder,recursive=True,return_relative_paths=True,convert_slashes=True)
image_files = find_images(input_base_folder,recursive=True,return_relative_paths=True,convert_slashes=True)

n_annotated_imgs = len([fn for fn in image_files if 'annotated-imgs' in fn])
print('Found {} images ({} in the annotated-imgs folder), {} videos'.format(
    len(image_files),n_annotated_imgs,len(video_files)))


#%% Read EXIF data

from megadetector.data_management.read_exif import read_exif_from_folder, ReadExifOptions

exif_options = ReadExifOptions()
exif_options.n_workers = 10

if os.path.isfile(exif_cache_file):
    print('EXIF cache {} exists, skipping EXIF read'.format(exif_cache_file))
    with open(exif_cache_file,'r') as f:
        exif_data = json.load(f)    
else:
    exif_data = read_exif_from_folder(input_folder=input_base_folder,
                                      output_file=exif_cache_file,
                                      options=exif_options,
                                      filenames=None,
                                      recursive=True)

assert len(exif_data) == len(image_files)


#%% Remove EXIF data

from megadetector.data_management.remove_exif import remove_exif
remove_exif(input_base_folder,recursive=True,n_processes=1)


#%% Read EXIF data again

exif_data_post_exif_removal = read_exif_from_folder(input_folder=input_base_folder,
                                                    output_file=exif_cache_file_post_exif_removal,
                                                    options=exif_options,
                                                    filenames=None,
                                                    recursive=True)


#%% Make sure no lat/lon data is present

from tqdm import tqdm

for i_image,im in enumerate(tqdm(exif_data_post_exif_removal)):
    tags = im['exif_tags']
    if tags is None:
        continue
    for k in tags:
        assert 'gps' not in str(k).lower()


#%% Look for images that contain humans

with open(md_results_file,'r') as f:
    md_results = json.load(f)

assert len(md_results['images']) == len(image_files)

human_threshold = 0.1
human_categories = ['2','3']

candidate_human_images = set()
failed_images = set()

# i_image = 0; im = md_results['images'][0]
for i_image,im in tqdm(enumerate(md_results['images']),total=len(md_results['images'])):

    if 'failure' in im:
        failed_images.add(im['file'])
        continue
    
    for det in im['detections']:
        if det['category'] in human_categories and det['conf'] >= human_threshold:
            candidate_human_images.add(im['file'])
            break
        
    # ...for each detection
    
# ...for each image    

print('Found {} failed images and {} candidate human images'.format(
    len(failed_images),len(candidate_human_images)))


#%% Copy failed images and human images to a temporary folder for review

review_folder_base = r'g:/temp/review_images'
os.makedirs(review_folder_base,exist_ok=True)

images_to_review = failed_images.union(candidate_human_images)
images_to_review = sorted(list(images_to_review))

source_file_to_target_file = {}

# fn_relative = images_to_review[0]
for fn_relative in images_to_review:
    assert '\\' not in fn_relative
    fn_abs_source = input_base_folder + '/' + fn_relative
    assert os.path.isfile(fn_abs_source)
    fn_abs_dest = review_folder_base + '/' + fn_relative.replace('/','_')
    source_file_to_target_file[fn_abs_source] = fn_abs_dest

from megadetector.utils.path_utils import parallel_copy_files

parallel_copy_files(input_file_to_output_file=source_file_to_target_file,
                    max_workers=16, 
                    use_threads=True,
                    overwrite=False,verbose=False)


#%% Copy videos to a temporary folder for review

review_folder_base = r'g:/temp/review_videos'
os.makedirs(review_folder_base,exist_ok=True)

source_file_to_target_file = {}

# fn_relative = video_files[0]
for fn_relative in video_files:
    assert '\\' not in fn_relative
    fn_abs_source = input_base_folder + '/' + fn_relative
    assert os.path.isfile(fn_abs_source)
    fn_abs_dest = review_folder_base + '/' + fn_relative.replace('/','_')
    source_file_to_target_file[fn_abs_source] = fn_abs_dest

from megadetector.utils.path_utils import parallel_copy_files

parallel_copy_files(input_file_to_output_file=source_file_to_target_file,
                    max_workers=16, 
                    use_threads=True,
                    overwrite=False,verbose=False)


#%% Track removed images

removed_images = [
    "annotated-imgs\panthera leo\Camera Trap\Events\X73Okngwe\2013\02\PvL_seq_41468415-4518-44d6-acac-2113b442f723\PICT0190.JPG",
    "annotated-imgs\panthera leo\Camera Trap\Hoanib\FldPln_Arch\211011\PvL_seq_5a9c6379-6980-4ab8-903a-b3bcba2ad21b\PICT0039.JPG",
    "annotated-imgs\panthera leo\Camera Trap\Hoanib\FldPln_Arch\211011\PvL_seq_5a9c6379-6980-4ab8-903a-b3bcba2ad21b\PICT0037.JPG",
    "annotated-imgs\panthera leo\Camera Trap\Hoanib\FldPln_Arch\211011\PvL_seq_5a9c6379-6980-4ab8-903a-b3bcba2ad21b\PICT0038.JPG",
    "annotated-imgs\panthera leo\Camera Trap\2015\09\PvL_seq_da9c9ab1-74a2-485e-b6e7-3827b0c2a2f0\20150924-RCX_0835.JPG",
    "annotated-imgs\panthera leo\Camera Trap\2015\09\PvL_seq_b0c1c6c5-474e-4844-a66c-e2bf5513d47a\20150924-RCX_0841.JPG",
    "annotated-imgs\oryx gazella\Camera Trap\Video_Clips\Leylands\CDY_0003.AVI"
]

removed_images = [fn.replace('\\','/') for fn in removed_images]


#%% Map filenames to datetimes

filename_to_datetime = {}
n_valid_datetimes = 0

# im = exif_data[0]
for im in tqdm(exif_data):
    if im['exif_tags'] is None or len(im['exif_tags']) == 0:
        filename_to_datetime[im['file_name']] = None
        continue
    dt = im['exif_tags']['DateTime']
    assert len(dt) == 19
    filename_to_datetime[im['file_name']] = dt
    n_valid_datetimes += 1

print('\nFound datetime information for {} of {} images'.format(
    n_valid_datetimes,len(exif_data)))    


#%% Convert "annotated_imgs" folder to COCO Camera Traps

from megadetector.utils.path_utils import recursive_file_list

species_name_to_category_id = {}

filenames_relative = \
    recursive_file_list(export_base,return_relative_paths=True,recursive=True,convert_slashes=True)
    
short_species_names = ['aves','cn-owls','cn-francolins','cn-raptors',
                       'columbidae','equus zebra hartmannae','numididae',
                       'pteroclidae']

images = []
annotations = []
n_datetimes = 0

for fn in filenames_relative:
    
    assert fn.lower().endswith('.jpg') or fn.lower().endswith('.avi') or fn.lower().endswith('.json')
    
    if fn.lower().endswith('.json'):
        continue

    tokens = fn.split('/')
    species_name = tokens[0]
    assert species_name in short_species_names or len(species_name.split(' ')) == 2
    
    if species_name not in species_name_to_category_id:
        category_id = len(species_name_to_category_id)
        species_name_to_category_id[species_name] = category_id
    else:
        category_id = species_name_to_category_id[species_name]
                
    im = {}
    im['id'] = fn
    im['file_name'] = fn
    im['location'] = 'unknown'
    
    fn_for_datetime_lookup = 'annotated-imgs/' + fn
    if (fn_for_datetime_lookup in filename_to_datetime) and \
        (filename_to_datetime[fn_for_datetime_lookup] is not None):
        im['datetime'] = filename_to_datetime[fn_for_datetime_lookup]
        n_datetimes += 1

    ann = {}
    ann['image_id'] = im['id']
    ann['id'] = im['id'] + ':ann_00'
    ann['sequence_level_annotation'] = False
    ann['category_id'] = category_id
    
    images.append(im)
    annotations.append(ann)
    
# ...for each filename

categories = []
for species_name in species_name_to_category_id:
    category = {}
    category['name'] = species_name
    category['id'] = species_name_to_category_id[species_name]
    categories.append(category)

info = {}
info['version'] = '2024.07.15_00'
info['description'] = 'Desert Lion Camera Traps'

d = {}
d['info'] = info
d['images'] = images
d['annotations'] = annotations
d['categories'] = categories

with open(output_file,'w') as f:
    json.dump(d,f,indent=1)


#%% Integrity check

from megadetector.data_management.databases.integrity_check_json_db import \
    IntegrityCheckOptions, integrity_check_json_db
    
integrity_check_options = IntegrityCheckOptions()

integrity_check_options.baseDir = export_base
integrity_check_options.bCheckImageExistence = True
integrity_check_options.bRequireLocation = True
integrity_check_options.nThreads = 10
integrity_check_options.verbose = True
integrity_check_options.allowIntIDs = False

integrity_check_results = integrity_check_json_db(output_file,integrity_check_options)


#%% Preview    

from megadetector.visualization.visualize_db \
    import DbVizOptions, visualize_db
    
viz_options = DbVizOptions()
viz_options.num_to_visualize = 2500

html_output_file,_ = visualize_db(output_file, preview_dir, export_base, options=viz_options)

from megadetector.utils.path_utils import open_file
open_file(html_output_file)


#%% Make MD results paths line up with the output

md_results_remapped_file = md_results_file.replace('-all','')
assert md_results_remapped_file != md_results_file

with open(output_file,'r') as f:
    d = json.load(f)
    
image_filenames = [im['file_name'] for im in d['images']]
image_filenames_set = set(image_filenames)

with open(md_results_file,'r') as f:
    md_results = json.load(f)

md_results_images_remapped = []

# im = md_results['images'][0]
for im in md_results['images']:
    assert im['file'].startswith('annotated-imgs/') or im['file'].startswith('bboxes/')
    if im['file'].startswith('bboxes/'):
        continue
    im['file'] = im['file'].replace('annotated-imgs/','')
    md_results_images_remapped.append(im)
    
print('Keeping {} of {} images in MD results'.format(
    len(md_results_images_remapped),len(md_results['images'])))

d['images'] = md_results_images_remapped

with open(md_results_remapped_file,'w') as f:
    json.dump(d,f,indent=1)


#%% Zip MD results and COCO file

from megadetector.utils.path_utils import zip_file

zip_file(input_fn=md_results_remapped_file, output_fn=None, overwrite=True, verbose=True, compresslevel=9)
zip_file(input_fn=output_file, output_fn=None, overwrite=True, verbose=True, compresslevel=9)


#%% Zip images

from megadetector.utils.path_utils import zip_folder

zip_folder(input_folder=export_base, output_fn=output_zipfile, overwrite=True, verbose=True, compresslevel=0)


#%% Copy lion images to a folder for thumbnail selection

review_folder_base = r'g:/temp/thumbnail-candidates'
os.makedirs(review_folder_base,exist_ok=True)

source_file_to_target_file = {}

# fn_relative = image_files[0]
for fn_relative in image_files:
    assert '\\' not in fn_relative
    if '/lion/' not in fn_relative and '/panthera leo/' not in fn_relative:
        continue
    fn_abs_source = input_base_folder + '/' + fn_relative
    assert os.path.isfile(fn_abs_source)
    fn_abs_dest = review_folder_base + '/' + fn_relative.replace('/','_')
    source_file_to_target_file[fn_abs_source] = fn_abs_dest

from megadetector.utils.path_utils import parallel_copy_files

parallel_copy_files(input_file_to_output_file=source_file_to_target_file,
                    max_workers=16, 
                    use_threads=True,
                    overwrite=False,verbose=False)



"""

video_utils.py

Utilities for splitting, rendering, and assembling videos.

"""

#%% Constants, imports, environment

import os
import re
import cv2
import glob
import json

from collections import defaultdict
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from tqdm import tqdm
from functools import partial
from inspect import signature

from megadetector.utils import path_utils    
from megadetector.utils.ct_utils import sort_list_of_dicts_by_key
from megadetector.visualization import visualization_utils as vis_utils

default_fourcc = 'h264'


#%% Path utilities

VIDEO_EXTENSIONS = ('.mp4','.avi','.mpeg','.mpg')

def is_video_file(s,video_extensions=VIDEO_EXTENSIONS):
    """
    Checks a file's extension against a set of known video file
    extensions to determine whether it's a video file.  Performs a
    case-insensitive comparison.
    
    Args:
        s (str): filename to check for probable video-ness
        video_extensions (list, optional): list of video file extensions
    
    Returns:
        bool: True if this looks like a video file, else False
    """
    
    ext = os.path.splitext(s)[1]
    return ext.lower() in video_extensions


def find_video_strings(strings):
    """
    Given a list of strings that are potentially video file names, looks for
    strings that actually look like video file names (based on extension).
    
    Args:
        strings (list): list of strings to check for video-ness
    
    Returns:
        list: a subset of [strings] that looks like they are video filenames
    """
    
    return [s for s in strings if is_video_file(s.lower())]


def find_videos(dirname, 
                recursive=False,
                convert_slashes=True,
                return_relative_paths=False):
    """
    Finds all files in a directory that look like video file names.
    
    Args:
        dirname (str): folder to search for video files
        recursive (bool, optional): whether to search [dirname] recursively
        convert_slashes (bool, optional): forces forward slashes in the returned files,
            otherwise uses the native path separator
        return_relative_paths (bool, optional): forces the returned filenames to be 
            relative to [dirname], otherwise returns absolute paths
    
    Returns:
        A list of filenames within [dirname] that appear to be videos
    """
    
    if recursive:
        files = glob.glob(os.path.join(dirname, '**', '*.*'), recursive=True)
    else:
        files = glob.glob(os.path.join(dirname, '*.*'))
        
    files = [fn for fn in files if os.path.isfile(fn)]
    
    if return_relative_paths:
        files = [os.path.relpath(fn,dirname) for fn in files]

    if convert_slashes:
        files = [fn.replace('\\', '/') for fn in files]
    
    return find_video_strings(files)


#%% Functions for rendering frames to video and vice-versa

# http://tsaith.github.io/combine-images-into-a-video-with-python-3-and-opencv-3.html

def frames_to_video(images, Fs, output_file_name, codec_spec=default_fourcc):
    """
    Given a list of image files and a sample rate, concatenates those images into
    a video and writes to a new video file.
    
    Args:
        images (list): a list of frame file names to concatenate into a video
        Fs (float): the frame rate in fps
        output_file_name (str): the output video file, no checking is performed to make
            sure the extension is compatible with the codec
        codec_spec (str, optional):  codec to use for encoding; h264 is a sensible default 
            and generally works on Windows, but when this fails (which is around 50% of the time 
            on Linux), mp4v is a good second choice
    """
    
    if codec_spec is None:
        codec_spec = 'h264'
        
    if len(images) == 0:
        print('Warning: no frames to render')
        return

    os.makedirs(os.path.dirname(output_file_name),exist_ok=True)
    
    # Determine the width and height from the first image
    frame = cv2.imread(images[0])
    cv2.imshow('video',frame)
    height, width, channels = frame.shape

    # Define the codec and create VideoWriter object
    fourcc = cv2.VideoWriter_fourcc(*codec_spec)
    out = cv2.VideoWriter(output_file_name, fourcc, Fs, (width, height))

    for image in images:
        frame = cv2.imread(image)
        out.write(frame)

    out.release()
    cv2.destroyAllWindows()


def get_video_fs(input_video_file):
    """
    Retrieves the frame rate of [input_video_file].
    
    Args:
        input_video_file (str): video file for which we want the frame rate
        
    Returns:
        float: the frame rate of [input_video_file]
    """
    
    assert os.path.isfile(input_video_file), 'File {} not found'.format(input_video_file)    
    vidcap = cv2.VideoCapture(input_video_file)
    Fs = vidcap.get(cv2.CAP_PROP_FPS)
    vidcap.release()
    return Fs


def _frame_number_to_filename(frame_number):
    """
    Ensures that frame images are given consistent filenames.
    """
    
    return 'frame{:06d}.jpg'.format(frame_number)


def _filename_to_frame_number(filename):
    """
    Extract the frame number from a filename that was created using 
    _frame_number_to_filename.
    
    Args:
        filename (str): a filename created with _frame_number_to_filename.
    Returns:
        int: the frame number extracted from [filename]
    """
    
    filename = os.path.basename(filename)
    match = re.search(r'frame(\d+)\.jpg', filename)
    if match is None:
        raise ValueError('{} does not appear to be a frame file'.format(filename))
    frame_number = match.group(1)
    try:
        frame_number = int(frame_number)
    except:
        raise ValueError('Filename {} does contain a valid frame number'.format(filename))
        
    return frame_number


def _add_frame_numbers_to_results(results):
    """
    Given the 'images' list from a set of MD results that was generated on video frames,
    add a 'frame_number' field to each image, and return the list, sorted by frame number.
    
    Args:
        results (list): list of image dicts        
    """
    
    # Add video-specific fields to the results
    for im in results:
        fn = im['file']
        frame_number = _filename_to_frame_number(fn)
        im['frame_number'] = frame_number
        
    results = sort_list_of_dicts_by_key(results,'frame_number')
    return results
    

def run_callback_on_frames(input_video_file, 
                           frame_callback,
                           every_n_frames=None, 
                           verbose=False, 
                           frames_to_process=None,
                           allow_empty_videos=False):
    """
    Calls the function frame_callback(np.array,image_id) on all (or selected) frames in
    [input_video_file].
    
    Args:
        input_video_file (str): video file to process
        frame_callback (function): callback to run on frames, should take an np.array and a string and 
            return a single value.  callback should expect PIL-formatted (RGB) images.
        every_n_frames (int, optional): sample every Nth frame starting from the first frame;
            if this is None or 1, every frame is processed. Mutually exclusive with
            frames_to_process.
        verbose (bool, optional): enable additional debug console output
        frames_to_process (list of int, optional): process this specific set of frames;
            mutually exclusive with every_n_frames.  If all values are beyond the length
            of the video, no frames are extracted.  Can also be a single int, specifying
            a single frame number.
        allow_empty_videos (bool, optional): Just print a warning if a video appears to have no
            frames (by default, this is an error).
    
    Returns:
        dict: dict with keys 'frame_filenames' (list), 'frame_rate' (float), 'results' (list).
        'frame_filenames' are synthetic filenames (e.g. frame000000.jpg); 'results' are
        in the same format used in the 'images' array in the MD results format.
    """
    
    assert os.path.isfile(input_video_file), 'File {} not found'.format(input_video_file)
    
    if isinstance(frames_to_process,int):
        frames_to_process = [frames_to_process]
        
    if (frames_to_process is not None) and (every_n_frames is not None):
        raise ValueError('frames_to_process and every_n_frames are mutually exclusive')
    
    vidcap = cv2.VideoCapture(input_video_file)
    n_frames = int(vidcap.get(cv2.CAP_PROP_FRAME_COUNT))
    frame_rate = vidcap.get(cv2.CAP_PROP_FPS)
    
    if verbose:
        print('Video {} contains {} frames at {} Hz'.format(input_video_file,n_frames,frame_rate))

    frame_filenames = []
    results = []
    
    # frame_number = 0
    for frame_number in range(0,n_frames):

        success,image = vidcap.read()
        
        if not success:
            assert image is None
            if verbose:
                print('Read terminating at frame {} of {}'.format(frame_number,n_frames))
            break

        if every_n_frames is not None:
            if frame_number % every_n_frames != 0:
                continue

        if frames_to_process is not None:
            if frame_number > max(frames_to_process):
                break
            if frame_number not in frames_to_process:
                continue
            
        frame_filename_relative = _frame_number_to_filename(frame_number)        
        frame_filenames.append(frame_filename_relative)
        
        image_np = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)        
        frame_results = frame_callback(image_np,frame_filename_relative)
        results.append(frame_results)
            
    # ...for each frame    
    
    if len(frame_filenames) == 0:
        if allow_empty_videos:
            print('Warning: found no frames in file {}'.format(input_video_file))
        else:
            raise Exception('Error: found no frames in file {}'.format(input_video_file))
        
    if verbose:
        print('\nProcessed {} of {} frames for {}'.format(
            len(frame_filenames),n_frames,input_video_file))

    vidcap.release()    
    to_return = {}
    to_return['frame_filenames'] = frame_filenames
    to_return['frame_rate'] = frame_rate
    to_return['results'] = results
    
    return to_return

# ...def run_callback_on_frames(...)


def run_callback_on_frames_for_folder(input_video_folder, 
                                      frame_callback,
                                      every_n_frames=None, 
                                      verbose=False,                                       
                                      allow_empty_videos=False,
                                      recursive=True):
    """
    Calls the function frame_callback(np.array,image_id) on all (or selected) frames in 
    all videos in [input_video_folder].
    
    Args:
        input_video_folder (str): video folder to process
        frame_callback (function): callback to run on frames, should take an np.array and a string and 
            return a single value.  callback should expect PIL-formatted (RGB) images.
        every_n_frames (int, optional): sample every Nth frame starting from the first frame;
            if this is None or 1, every frame is processed.
        verbose (bool, optional): enable additional debug console output
        allow_empty_videos (bool, optional): Just print a warning if a video appears to have no
            frames (by default, this is an error).
        recursive (bool, optional): recurse into [input_video_folder]
    
    Returns:
        dict: dict with keys 'video_filenames' (list of str), 'frame_rates' (list of floats),
        'results' (list of list of dicts). 'video_filenames' will contain *relative* filenames.
    """
    
    to_return = {'video_filenames':[],'frame_rates':[],'results':[]}
    
    # Recursively enumerate video files
    input_files_full_paths = find_videos(input_video_folder,
                                         recursive=recursive,
                                         convert_slashes=True,
                                         return_relative_paths=False)
    print('Found {} videos in folder {}'.format(len(input_files_full_paths),input_video_folder))
    
    if len(input_files_full_paths) == 0:
        return to_return
        
    # Process each video
    
    # video_fn_abs = input_files_full_paths[0]
    for video_fn_abs in tqdm(input_files_full_paths):
        video_results = run_callback_on_frames(input_video_file=video_fn_abs,
                                               frame_callback=frame_callback,
                                               every_n_frames=every_n_frames, 
                                               verbose=verbose, 
                                               frames_to_process=None,
                                               allow_empty_videos=allow_empty_videos)
        
        """
        dict: dict with keys 'frame_filenames' (list), 'frame_rate' (float), 'results' (list).
            'frame_filenames' are synthetic filenames (e.g. frame000000.jpg); 'results' are
            in the same format used in the 'images' array in the MD results format.
        """
        video_filename_relative = os.path.relpath(video_fn_abs,input_video_folder)
        video_filename_relative = video_filename_relative.replace('\\','/')
        to_return['video_filenames'].append(video_filename_relative)
        to_return['frame_rates'].append(video_results['frame_rate'])
        for r in video_results['results']:
            assert r['file'].startswith('frame')
            r['file'] = video_filename_relative + '/' + r['file']
        to_return['results'].append(video_results['results'])
        
    # ...for each video
    
    n_videos = len(input_files_full_paths)
    assert len(to_return['video_filenames']) == n_videos
    assert len(to_return['frame_rates']) == n_videos
    assert len(to_return['results']) == n_videos
    
    return to_return

# ...def run_callback_on_frames_for_folder(...)

    
def video_to_frames(input_video_file, 
                    output_folder, 
                    overwrite=True, 
                    every_n_frames=None, 
                    verbose=False, 
                    quality=None,
                    max_width=None, 
                    frames_to_extract=None,
                    allow_empty_videos=False):
    """
    Renders frames from [input_video_file] to .jpg files in [output_folder].
    
    With help from:
        
    https://stackoverflow.com/questions/33311153/python-extracting-and-saving-video-frames
    
    Args:
        input_video_file (str): video file to split into frames
        output_folder (str): folder to put frame images in
        overwrite (bool, optional): whether to overwrite existing frame images
        every_n_frames (int, optional): sample every Nth frame starting from the first frame;
            if this is None or 1, every frame is extracted. Mutually exclusive with
            frames_to_extract.
        verbose (bool, optional): enable additional debug console output
        quality (int, optional): JPEG quality for frame output, from 0-100.  Defaults
            to the opencv default (typically 95).
        max_width (int, optional): resize frames to be no wider than [max_width]
        frames_to_extract (list of int, optional): extract this specific set of frames;
            mutually exclusive with every_n_frames.  If all values are beyond the length
            of the video, no frames are extracted.  Can also be a single int, specifying
            a single frame number.
        allow_empty_videos (bool, optional): Just print a warning if a video appears to have no
            frames (by default, this is an error).
    
    Returns:
        tuple: length-2 tuple containing (list of frame filenames,frame rate)
    """
    
    assert os.path.isfile(input_video_file), 'File {} not found'.format(input_video_file)
    
    if quality is not None and quality < 0:
        quality = None
        
    if isinstance(frames_to_extract,int):
        frames_to_extract = [frames_to_extract]
        
    if (frames_to_extract is not None) and (every_n_frames is not None):
        raise ValueError('frames_to_extract and every_n_frames are mutually exclusive')
    
    os.makedirs(output_folder,exist_ok=True)
    
    vidcap = cv2.VideoCapture(input_video_file)
    n_frames = int(vidcap.get(cv2.CAP_PROP_FRAME_COUNT))
    Fs = vidcap.get(cv2.CAP_PROP_FPS)
    
    # If we're not over-writing, check whether all frame images already exist
    if overwrite == False:
        
        missing_frame_number = None
        missing_frame_filename = None
        frame_filenames = []
        found_existing_frame = False
        
        for frame_number in range(0,n_frames):
            
            if every_n_frames is not None:
                assert frames_to_extract is None, \
                    'Internal error: frames_to_extract and every_n_frames are exclusive'
                if (frame_number % every_n_frames) != 0:
                    continue
            
            if frames_to_extract is not None:
                assert every_n_frames is None, \
                    'Internal error: frames_to_extract and every_n_frames are exclusive'
                if frame_number not in frames_to_extract:
                    continue
                
            frame_filename = _frame_number_to_filename(frame_number)
            frame_filename = os.path.join(output_folder,frame_filename)
            frame_filenames.append(frame_filename)
            if os.path.isfile(frame_filename):
                found_existing_frame = True
                continue
            else:
                missing_frame_number = frame_number
                missing_frame_filename = frame_filename
                break
    
        if verbose and missing_frame_number is not None:
            print('Missing frame {} ({}) for video {}'.format(
                missing_frame_number,
                missing_frame_filename,
                input_video_file))
            
        # OpenCV seems to over-report the number of frames by 1 in some cases, or fails
        # to read the last frame; either way, I'm allowing one missing frame.
        allow_last_frame_missing = True
        
        # This doesn't have to mean literally the last frame number, it just means that if
        # we find this frame or later, we consider the video done
        last_expected_frame_number = n_frames-1
        if every_n_frames is not None:
            last_expected_frame_number -= (every_n_frames*2)
          
        # When specific frames are requested, if anything is missing, reprocess the video
        if (frames_to_extract is not None) and (missing_frame_number is not None):
            
            pass
        
        # If no frames are missing, or only frames very close to the end of the video are "missing",
        # skip this video
        elif (missing_frame_number is None) or \
            (allow_last_frame_missing and (missing_frame_number >= last_expected_frame_number)):
                
            if verbose:
                print('Skipping video {}, all output frames exist'.format(input_video_file))
            return frame_filenames,Fs
        
        else:
            
            # If we found some frames, but not all, print a message
            if verbose and found_existing_frame:
                print("Rendering video {}, couldn't find frame {} ({}) of {}".format(
                    input_video_file,
                    missing_frame_number,
                    missing_frame_filename,
                    last_expected_frame_number))
    
    # ...if we need to check whether to skip this video entirely
        
    if verbose:
        print('Video {} contains {} frames at {} Hz'.format(input_video_file,n_frames,Fs))

    frame_filenames = []

    # YOLOv5 does some totally bananas monkey-patching of opencv, which causes 
    # problems  if we try to supply a third parameter to imwrite (to specify JPEG 
    # quality).  Detect  this case, and ignore the quality  parameter if it looks 
    # like imwrite has been messed with.
    #
    # See:
    #
    # https://github.com/ultralytics/yolov5/issues/7285
    imwrite_patched = False
    n_imwrite_parameters = None
    
    try:
        # calling signature() on the native cv2.imwrite function will
        # fail, so an exception here is a good thing.  In fact I don't think
        # there's a case where this *succeeds* and the number of parameters
        # is wrong.
        sig = signature(cv2.imwrite)
        n_imwrite_parameters = len(sig.parameters)
    except Exception:
        pass
    
    if (n_imwrite_parameters is not None) and (n_imwrite_parameters < 3):
        imwrite_patched = True
        if verbose and (quality is not None):
            print('Warning: quality value supplied, but YOLOv5 has mucked with cv2.imwrite, ignoring quality')
            
    # for frame_number in tqdm(range(0,n_frames)):
    for frame_number in range(0,n_frames):

        success,image = vidcap.read()
        if not success:
            assert image is None
            if verbose:
                print('Read terminating at frame {} of {}'.format(frame_number,n_frames))
            break

        if every_n_frames is not None:
            if frame_number % every_n_frames != 0:
                continue

        if frames_to_extract is not None:
            if frame_number > max(frames_to_extract):
                break
            if frame_number not in frames_to_extract:
                continue
            
        # Has resizing been requested?
        if max_width is not None:
        
            # image.shape is h/w/dims            
            input_shape = image.shape
            assert input_shape[2] == 3
            input_width = input_shape[1]
            
            # Is resizing necessary?
            if input_width > max_width:
                
                scale = max_width / input_width
                assert scale <= 1.0
                
                # INTER_AREA is recommended for size reduction
                image = cv2.resize(image, (0,0), fx=scale, fy=scale, interpolation=cv2.INTER_AREA)
        
        # ...if we need to deal with resizing
        
        frame_filename_relative = _frame_number_to_filename(frame_number)        
        frame_filename = os.path.join(output_folder,frame_filename_relative)
        frame_filenames.append(frame_filename)
        
        if overwrite == False and os.path.isfile(frame_filename):
            # print('Skipping frame {}'.format(frame_filename))
            pass            
        else:
            try:
                if frame_filename.isascii():
                    
                    if quality is None or imwrite_patched:
                        cv2.imwrite(os.path.normpath(frame_filename),image)
                    else:                        
                        cv2.imwrite(os.path.normpath(frame_filename),image,
                                    [int(cv2.IMWRITE_JPEG_QUALITY), quality])
                else:
                    if quality is None:
                        is_success, im_buf_arr = cv2.imencode('.jpg', image)
                    else:
                        encode_param = [int(cv2.IMWRITE_JPEG_QUALITY), quality]
                        is_success, im_buf_arr = cv2.imencode('.jpg', image, encode_param)
                    im_buf_arr.tofile(frame_filename)
                assert os.path.isfile(frame_filename), \
                    'Output frame {} unavailable'.format(frame_filename)
            except KeyboardInterrupt:
                vidcap.release()
                raise
            except Exception as e:
                print('Error on frame {} of {}: {}'.format(frame_number,n_frames,str(e)))

    # ...for each frame
    
    if len(frame_filenames) == 0:
        if allow_empty_videos:
            print('Warning: found no frames in file {}'.format(input_video_file))
        else:
            raise Exception('Error: found no frames in file {}'.format(input_video_file))
        
    if verbose:
        print('\nExtracted {} of {} frames for {}'.format(
            len(frame_filenames),n_frames,input_video_file))

    vidcap.release()    
    return frame_filenames,Fs

# ...def video_to_frames(...)


def _video_to_frames_for_folder(relative_fn,input_folder,output_folder_base,
                                every_n_frames,overwrite,verbose,quality,max_width,
                                frames_to_extract,allow_empty_videos):
    """
    Internal function to call video_to_frames for a single video in the context of 
    video_folder_to_frames;  makes sure the right output folder exists, then calls 
    video_to_frames.
    """    
    
    input_fn_absolute = os.path.join(input_folder,relative_fn)
    assert os.path.isfile(input_fn_absolute),\
        'Could not find file {}'.format(input_fn_absolute)

    # Create the target output folder
    output_folder_video = os.path.join(output_folder_base,relative_fn)
    os.makedirs(output_folder_video,exist_ok=True)

    # Render frames
    # input_video_file = input_fn_absolute; output_folder = output_folder_video
    frame_filenames,fs = video_to_frames(input_fn_absolute,
                                         output_folder_video,
                                         overwrite=overwrite,
                                         every_n_frames=every_n_frames,
                                         verbose=verbose,
                                         quality=quality,
                                         max_width=max_width,
                                         frames_to_extract=frames_to_extract,
                                         allow_empty_videos=allow_empty_videos)
    
    return frame_filenames,fs


def video_folder_to_frames(input_folder, output_folder_base, 
                           recursive=True, overwrite=True,
                           n_threads=1, every_n_frames=None,
                           verbose=False, parallelization_uses_threads=True,
                           quality=None, max_width=None, 
                           frames_to_extract=None, allow_empty_videos=False):
    """
    For every video file in input_folder, creates a folder within output_folder_base, and 
    renders frame of that video to images in that folder.
    
    Args:
        input_folder (str): folder to process
        output_folder_base (str): root folder for output images; subfolders will be
            created for each input video
        recursive (bool, optional): whether to recursively process videos in [input_folder]
        overwrite (bool, optional): whether to overwrite existing frame images
        n_threads (int, optional): number of concurrent workers to use; set to <= 1 to disable
            parallelism
        every_n_frames (int, optional): sample every Nth frame starting from the first frame;
            if this is None or 1, every frame is extracted. Mutually exclusive with 
            frames_to_extract.
        verbose (bool, optional): enable additional debug console output
        parallelization_uses_threads (bool, optional): whether to use threads (True) or
            processes (False) for parallelization; ignored if n_threads <= 1
        quality (int, optional): JPEG quality for frame output, from 0-100.  Defaults
            to the opencv default (typically 95).
        max_width (int, optional): resize frames to be no wider than [max_width]
        frames_to_extract (list of int, optional): extract this specific set of frames from
            each video; mutually exclusive with every_n_frames.  If all values are beyond 
            the length of a video, no frames are extracted. Can also be a single int, 
            specifying a single frame number.
        
    Returns:
        tuple: a length-3 tuple containing:
            - list of lists of frame filenames; the Nth list of frame filenames corresponds to 
              the Nth video
            - list of video frame rates; the Nth value corresponds to the Nth video
            - list of video filenames    
    """
    
    # Recursively enumerate video files
    input_files_full_paths = find_videos(input_folder,recursive=recursive)
    print('Found {} videos in folder {}'.format(len(input_files_full_paths),input_folder))
    if len(input_files_full_paths) == 0:
        return [],[],[]
    
    input_files_relative_paths = [os.path.relpath(s,input_folder) for s in input_files_full_paths]
    input_files_relative_paths = [s.replace('\\','/') for s in input_files_relative_paths]
    
    os.makedirs(output_folder_base,exist_ok=True)    
    
    frame_filenames_by_video = []
    fs_by_video = []
    
    if n_threads == 1:
        # For each video
        #
        # input_fn_relative = input_files_relative_paths[0]
        for input_fn_relative in tqdm(input_files_relative_paths):
        
            frame_filenames,fs = \
                _video_to_frames_for_folder(input_fn_relative,input_folder,output_folder_base,
                                            every_n_frames,overwrite,verbose,quality,max_width,
                                            frames_to_extract,allow_empty_videos)
            frame_filenames_by_video.append(frame_filenames)
            fs_by_video.append(fs)
    else:
        if parallelization_uses_threads:
            print('Starting a worker pool with {} threads'.format(n_threads))
            pool = ThreadPool(n_threads)
        else:
            print('Starting a worker pool with {} processes'.format(n_threads))
            pool = Pool(n_threads)
        process_video_with_options = partial(_video_to_frames_for_folder, 
                                             input_folder=input_folder,
                                             output_folder_base=output_folder_base,
                                             every_n_frames=every_n_frames,
                                             overwrite=overwrite,
                                             verbose=verbose,
                                             quality=quality,
                                             max_width=max_width,
                                             frames_to_extract=frames_to_extract,
                                             allow_empty_videos=allow_empty_videos)
        results = list(tqdm(pool.imap(
            partial(process_video_with_options),input_files_relative_paths), 
                            total=len(input_files_relative_paths)))
        frame_filenames_by_video = [x[0] for x in results]
        fs_by_video = [x[1] for x in results]
        
    return frame_filenames_by_video,fs_by_video,input_files_full_paths
  
# ...def video_folder_to_frames(...)


class FrameToVideoOptions:
    """
    Options controlling the conversion of frame-level results to video-level results via
    frame_results_to_video_results()    
    """
    
    def __init__(self):
            
        #: One-indexed indicator of which frame-level confidence value to use to determine detection confidence
        #: for the whole video, i.e. "1" means "use the confidence value from the highest-confidence frame"
        self.nth_highest_confidence = 1
        
        #: What to do if a file referred to in a .json results file appears not to be a 
        #: video; can be 'error' or 'skip_with_warning'
        self.non_video_behavior = 'error'
    

def frame_results_to_video_results(input_file,output_file,options=None,
                                   video_filename_to_frame_rate=None):
    """
    Given an MD results file produced at the *frame* level, corresponding to a directory 
    created with video_folder_to_frames, maps those frame-level results back to the 
    video level for use in Timelapse.
    
    Preserves everything in the input .json file other than the images.
    
    Args:
        input_file (str): the frame-level MD results file to convert to video-level results
        output_file (str): the .json file to which we should write video-level results
        options (FrameToVideoOptions, optional): parameters for converting frame-level results
            to video-level results, see FrameToVideoOptions for details            
        video_filename_to_frame_rate (dict): maps (relative) video path names to frame rates,
            used only to populate the output file
    """

    if options is None:
        options = FrameToVideoOptions()
        
    # Load results
    with open(input_file,'r') as f:
        input_data = json.load(f)

    images = input_data['images']
    detection_categories = input_data['detection_categories']
    
    
    ## Break into videos
    
    video_to_frame_info = defaultdict(list) 
    
    # im = images[0]
    for im in tqdm(images):
        
        fn = im['file']
        video_name = os.path.dirname(fn)
        
        if not is_video_file(video_name):
            
            if options.non_video_behavior == 'error':
                raise ValueError('{} is not a video file'.format(video_name))
            elif options.non_video_behavior == 'skip_with_warning':
                print('Warning: {} is not a video file'.format(video_name))
                continue
            else:
                raise ValueError('Unrecognized non-video handling behavior: {}'.format(
                    options.non_video_behavior))
        
        # Attach video-specific fields to the output, specifically attach the frame
        # number to both the video and each detection.  Only the frame number for the 
        # canonical detection will end up in the video-level output file.
        frame_number = _filename_to_frame_number(fn)
        im['frame_number'] = frame_number
        for detection in im['detections']: 
            detection['frame_number'] = frame_number
        
        video_to_frame_info[video_name].append(im)
    
    # ...for each frame referred to in the results file
    
    print('Found {} unique videos in {} frame-level results'.format(
        len(video_to_frame_info),len(images)))
    
    output_images = []
    
    
    ## For each video...
    
    # video_name = list(video_to_frame_info.keys())[0]
    for video_name in tqdm(video_to_frame_info):
        
        frames = video_to_frame_info[video_name]
        
        all_detections_this_video = []
        
        # frame = frames[0]
        for frame in frames:
            if ('detections' in frame) and (frame['detections'] is not None):
                all_detections_this_video.extend(frame['detections'])
            
        # At most one detection for each category for the whole video
        canonical_detections = []
            
        # category_id = list(detection_categories.keys())[0]
        for category_id in detection_categories:
            
            category_detections = [det for det in all_detections_this_video if \
                                   det['category'] == category_id]
            
            # Find the nth-highest-confidence video to choose a confidence value
            if len(category_detections) >= options.nth_highest_confidence:
                
                category_detections_by_confidence = sorted(category_detections, 
                                                           key = lambda i: i['conf'],reverse=True)
                canonical_detection = category_detections_by_confidence[options.nth_highest_confidence-1]
                canonical_detections.append(canonical_detection)
                                      
        # Prepare the output representation for this video
        im_out = {}
        im_out['file'] = video_name
        im_out['detections'] = canonical_detections
        
        if (video_filename_to_frame_rate is not None) and \
            (video_name in video_filename_to_frame_rate):
            im_out['frame_rate'] = video_filename_to_frame_rate[video_name]
        
        # 'max_detection_conf' is no longer included in output files by default
        if False:
            im_out['max_detection_conf'] = 0
            if len(canonical_detections) > 0:
                confidences = [d['conf'] for d in canonical_detections]
                im_out['max_detection_conf'] = max(confidences)
        
        output_images.append(im_out)
        
    # ...for each video
    
    output_data = input_data
    output_data['images'] = output_images
    s = json.dumps(output_data,indent=1)
    
    # Write the output file
    with open(output_file,'w') as f:
        f.write(s)
    
# ...def frame_results_to_video_results(...)


#%% Test drivers

if False:

    pass

    #%% Constants
    
    input_folder = r'G:\temp\usu-long\data'
    frame_folder_base = r'g:\temp\usu-long-single-frames'
    assert os.path.isdir(input_folder)
        
    
    #%% Split videos into frames
        
    frame_filenames_by_video,fs_by_video,video_filenames = \
        video_folder_to_frames(input_folder,
                               frame_folder_base,
                               recursive=True,
                               overwrite=True,
                               n_threads=10, 
                               every_n_frames=None,
                               verbose=True, 
                               parallelization_uses_threads=True,
                               quality=None, 
                               max_width=None, 
                               frames_to_extract=150)
    
    
    #%% Constants for detection tests
    
    detected_frame_folder_base = r'e:\video_test\detected_frames'
    rendered_videos_folder_base = r'e:\video_test\rendered_videos'
    os.makedirs(detected_frame_folder_base,exist_ok=True)
    os.makedirs(rendered_videos_folder_base,exist_ok=True)
    results_file = r'results.json'
    confidence_threshold = 0.75
        
    #%% Load detector output    
    
    with open(results_file,'r') as f:
        detection_results = json.load(f)
    detections = detection_results['images']
    detector_label_map = detection_results['detection_categories']
    for d in detections:
        d['file'] = d['file'].replace('\\','/').replace('video_frames/','')


    #%% List image files, break into folders
    
    frame_files = path_utils.find_images(frame_folder_base,True)
    frame_files = [s.replace('\\','/') for s in frame_files]
    print('Enumerated {} total frames'.format(len(frame_files)))
    
    # Find unique folders
    folders = set()
    # fn = frame_files[0]
    for fn in frame_files:
        folders.add(os.path.dirname(fn))
    folders = [s.replace('\\','/') for s in folders]
    print('Found {} folders for {} files'.format(len(folders),len(frame_files)))
    
        
    #%% Render detector frames
    
    # folder = list(folders)[0]
    for folder in folders:
        
        frame_files_this_folder = [fn for fn in frame_files if folder in fn]
        folder_relative = folder.replace((frame_folder_base + '/').replace('\\','/'),'')
        detection_results_this_folder = [d for d in detections if folder_relative in d['file']]
        print('Found {} detections in folder {}'.format(len(detection_results_this_folder),folder))
        assert len(frame_files_this_folder) == len(detection_results_this_folder)
        
        rendered_frame_output_folder = os.path.join(detected_frame_folder_base,folder_relative)
        os.makedirs(rendered_frame_output_folder,exist_ok=True)
        
        # d = detection_results_this_folder[0]
        for d in tqdm(detection_results_this_folder):
            
            input_file = os.path.join(frame_folder_base,d['file'])
            output_file = os.path.join(detected_frame_folder_base,d['file'])
            os.makedirs(os.path.dirname(output_file),exist_ok=True)
            vis_utils.draw_bounding_boxes_on_file(input_file,output_file,d['detections'],
                                                  confidence_threshold)
        
        # ...for each file in this folder
            
    # ...for each folder


    #%% Render output videos
            
    # folder = list(folders)[0]
    for folder in tqdm(folders):
        
        folder_relative = folder.replace((frame_folder_base + '/').replace('\\','/'),'')
        rendered_detector_output_folder = os.path.join(detected_frame_folder_base,folder_relative)
        assert os.path.isdir(rendered_detector_output_folder)
        
        frame_files_relative = os.listdir(rendered_detector_output_folder)
        frame_files_absolute = [os.path.join(rendered_detector_output_folder,s) \
                                for s in frame_files_relative]
        
        output_video_filename = os.path.join(rendered_videos_folder_base,folder_relative)
        os.makedirs(os.path.dirname(output_video_filename),exist_ok=True)
        
        original_video_filename = output_video_filename.replace(
            rendered_videos_folder_base,input_folder)
        assert os.path.isfile(original_video_filename)
        Fs = get_video_fs(original_video_filename)
                
        frames_to_video(frame_files_absolute, Fs, output_video_filename)

    # ...for each video



"""

pytorch_detector.py

Module to run MegaDetector v5.

"""

#%% Imports and constants

import torch
import numpy as np
import traceback

from megadetector.detection.run_detector import CONF_DIGITS, COORD_DIGITS, FAILURE_INFER
from megadetector.utils import ct_utils

# We support a few ways of accessing the YOLOv5 dependencies:
#
# * The standard configuration as of 2023.09 expects that the YOLOv5 repo is checked
#   out and on the PYTHONPATH (import utils)
#
# * Supported but non-default (used for PyPI packaging):
#
#   pip install ultralytics-yolov5
#
# * Works, but not supported:
#
#   pip install yolov5
#
# * Unfinished:
#
#   pip install ultralytics
#
#   If try_ultralytics_import is True, we'll try to import all YOLOv5 dependencies from 
#   ultralytics.utils and ultralytics.data.  But as of 2023.11, this results in a "No 
#   module named 'models'" error when running MDv5, and there's no upside to this approach
#   compared to using either of the YOLOv5 PyPI packages, so... punting on this for now.

utils_imported = False
try_yolov5_import = True

# See above; this should remain as "False" unless we update the MegaDetector .pt file
# to use more recent YOLOv5 namespace conventions.
try_ultralytics_import = False

# First try importing from the yolov5 package; this is how the pip
# package finds YOLOv5 utilities.
if try_yolov5_import and not utils_imported:
    
    try:
        from yolov5.utils.general import non_max_suppression, xyxy2xywh # noqa
        from yolov5.utils.augmentations import letterbox # noqa
        from yolov5.utils.general import scale_boxes as scale_coords # noqa
        utils_imported = True
        print('Imported YOLOv5 from YOLOv5 package')
    except Exception:
        # print('YOLOv5 module import failed, falling back to path-based import')
        pass

# If we haven't succeeded yet, import from the ultralytics package        
if try_ultralytics_import and not utils_imported:
    
    try:
        from ultralytics.utils.ops import non_max_suppression # noqa
        from ultralytics.utils.ops import xyxy2xywh # noqa
        from ultralytics.utils.ops import scale_coords # noqa
        from ultralytics.data.augment import LetterBox
        
        # letterbox() became a LetterBox class in the ultralytics package
        def letterbox(img,new_shape,stride,auto=True): # noqa
            L = LetterBox(new_shape,stride=stride,auto=auto)
            letterbox_result = L(image=img)
            return [letterbox_result]
        utils_imported = True
        print('Imported YOLOv5 from ultralytics package')
    except Exception:
        # print('Ultralytics module import failed, falling back to yolov5 import')
        pass

# If we haven't succeeded yet, assume the YOLOv5 repo is on our PYTHONPATH.
if not utils_imported:
    
    try:
        # import pre- and post-processing functions from the YOLOv5 repo
        from utils.general import non_max_suppression, xyxy2xywh # noqa
        from utils.augmentations import letterbox # noqa
        
        # scale_coords() became scale_boxes() in later YOLOv5 versions
        try:
            from utils.general import scale_coords # noqa
        except ImportError:
            from utils.general import scale_boxes as scale_coords
        utils_imported = True
        print('Imported YOLOv5 as utils.*')
                
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError('Could not import YOLOv5 functions:\n{}'.format(str(e)))

assert utils_imported, 'YOLOv5 import error'

print(f'Using PyTorch version {torch.__version__}')

    
#%% Classes

class PTDetector:

    #: Image size passed to YOLOv5's letterbox() function; 1280 means "1280 on the long side, preserving 
    #: aspect ratio"
    #:
    #: :meta private:
    IMAGE_SIZE = 1280
    
    #: Stride size passed to YOLOv5's letterbox() function
    #:
    #: :meta private:
    STRIDE = 64

    def __init__(self, model_path, force_cpu=False, use_model_native_classes= False):
        
        self.device = 'cpu'
        if not force_cpu:
            if torch.cuda.is_available():
                self.device = torch.device('cuda:0')
            try:
                if torch.backends.mps.is_built and torch.backends.mps.is_available():
                    self.device = 'mps'
            except AttributeError:
                pass
        try:
            self.model = PTDetector._load_model(model_path, self.device)
        except Exception as e:
            # In a very esoteric scenario where an old version of YOLOv5 is used to run
            # newer models, we run into an issue because the "Model" class became
            # "DetectionModel".  New YOLOv5 code handles this case by just setting them
            # to be the same, so doing that via monkey-patch doesn't seem *that* rude.
            if "Can't get attribute 'DetectionModel'" in str(e):
                print('Forward-compatibility issue detected, patching')
                from models import yolo
                yolo.DetectionModel = yolo.Model                
                self.model = PTDetector._load_model(model_path, self.device)                
            else:
                raise
        if (self.device != 'cpu'):
            print('Sending model to GPU')
            self.model.to(self.device)
            
        self.printed_image_size_warning = False        
        self.use_model_native_classes = use_model_native_classes
        

    @staticmethod
    def _load_model(model_pt_path, device):
        
        # There are two very slightly different ways to load the model, (1) using the
        # map_location=device parameter to torch.load and (2) calling .to(device) after
        # loading the model.  The former is what we did for a zillion years, but is not
        # supported on Apple silicon at of 2029.09.  Switching to the latter causes
        # very slight changes to the output, which always make me nervous, so I'm not
        # doing a wholesale swap just yet.  Instead, we'll just do this on M1 hardware.
        use_map_location = (device != 'mps')        
        
        if use_map_location:
            checkpoint = torch.load(model_pt_path, map_location=device)
        else:
            checkpoint = torch.load(model_pt_path)
        
        # Compatibility fix that allows us to load older YOLOv5 models with 
        # newer versions of YOLOv5/PT
        for m in checkpoint['model'].modules():
            t = type(m)
            if t is torch.nn.Upsample and not hasattr(m, 'recompute_scale_factor'):
                m.recompute_scale_factor = None
        
        if use_map_location:
            model = checkpoint['model'].float().fuse().eval()
        else:
            model = checkpoint['model'].float().fuse().eval().to(device)
            
        return model

    def generate_detections_one_image(self, 
                                      img_original, 
                                      image_id='unknown', 
                                      detection_threshold=0.00001, 
                                      image_size=None,
                                      skip_image_resizing=False,
                                      augment=False):
        """
        Applies the detector to an image.

        Args:
            img_original (Image): the PIL Image object (or numpy array) on which we should run the 
                detector, with EXIF rotation already handled.
            image_id (str, optional): a path to identify the image; will be in the "file" field 
                of the output object
            detection_threshold (float, optional): only detections above this confidence threshold 
                will be included in the return value
            image_size (tuple, optional): image size to use for inference, only mess with this if 
                (a) you're using a model other than MegaDetector or (b) you know what you're getting into
            skip_image_resizing (bool, optional): whether to skip internal image resizing (and rely on 
                external resizing)
            augment (bool, optional): enable (implementation-specific) image augmentation

        Returns:
            dict: a dictionary with the following fields:
                - 'file' (filename, always present)
                - 'max_detection_conf' (removed from MegaDetector output files by default, but generated here)
                - 'detections' (a list of detection objects containing keys 'category', 'conf', and 'bbox')
                - 'failure' (a failure string, or None if everything went fine)
        """

        result = {'file': image_id }
        detections = []
        max_conf = 0.0

        if detection_threshold is None:
            
            detection_threshold = 0
            
        try:
            
            if not isinstance(img_original,np.ndarray):                
                img_original = np.asarray(img_original)

            # Padded resize
            target_size = PTDetector.IMAGE_SIZE
            
            # Image size can be an int (which translates to a square target size) or (h,w)
            if image_size is not None:
                
                assert isinstance(image_size,int) or (len(image_size)==2)
                
                if not self.printed_image_size_warning:
                    print('Warning: using user-supplied image size {}'.format(image_size))
                    self.printed_image_size_warning = True
            
                target_size = image_size
            
            else:
                
                self.printed_image_size_warning = False
                
            # ...if the caller has specified an image size
            
            if skip_image_resizing:
                img = img_original
            else:
                letterbox_result = letterbox(img_original, 
                                             new_shape=target_size,
                                             stride=PTDetector.STRIDE, 
                                             auto=True)
                img = letterbox_result[0]                
            
            # HWC to CHW; PIL Image is RGB already
            img = img.transpose((2, 0, 1))
            img = np.ascontiguousarray(img)
            img = torch.from_numpy(img)
            img = img.to(self.device)
            img = img.float()
            img /= 255

            # In practice this is always true 
            if len(img.shape) == 3:  
                img = torch.unsqueeze(img, 0)

            pred = self.model(img,augment=augment)[0]

            # NMS
            if self.device == 'mps':
                # As of PyTorch 1.13.0.dev20220824, nms is not implemented for MPS.
                #
                # Send predictions back to the CPU for NMS.
                pred = non_max_suppression(prediction=pred.cpu(), conf_thres=detection_threshold)
            else: 
                pred = non_max_suppression(prediction=pred, conf_thres=detection_threshold)

            # format detections/bounding boxes
            #
            # normalization gain whwh
            gn = torch.tensor(img_original.shape)[[1, 0, 1, 0]]

            # This is a loop over detection batches, which will always be length 1 in our case,
            # since we're not doing batch inference.
            for det in pred:
                
                if len(det):
                    
                    # Rescale boxes from img_size to im0 size
                    det[:, :4] = scale_coords(img.shape[2:], det[:, :4], img_original.shape).round()

                    for *xyxy, conf, cls in reversed(det):
                        
                        # normalized center-x, center-y, width and height
                        xywh = (xyxy2xywh(torch.tensor(xyxy).view(1, 4)) / gn).view(-1).tolist()

                        api_box = ct_utils.convert_yolo_to_xywh(xywh)

                        conf = ct_utils.truncate_float(conf.tolist(), precision=CONF_DIGITS)

                        if not self.use_model_native_classes:
                            # MegaDetector output format's categories start at 1, but the MD 
                            # model's categories start at 0.
                            cls = int(cls.tolist()) + 1
                            if cls not in (1, 2, 3):
                                raise KeyError(f'{cls} is not a valid class.')
                        else:
                            cls = int(cls.tolist())

                        detections.append({
                            'category': str(cls),
                            'conf': conf,
                            'bbox': ct_utils.truncate_float_array(api_box, precision=COORD_DIGITS)
                        })
                        max_conf = max(max_conf, conf)
                        
                    # ...for each detection in this batch
                        
                # ...if this is a non-empty batch
                
            # ...for each detection batch

        # ...try
        
        except Exception as e:
            
            result['failure'] = FAILURE_INFER
            print('PTDetector: image {} failed during inference: {}\n'.format(image_id, str(e)))
            traceback.print_exc(e)

        result['max_detection_conf'] = max_conf
        result['detections'] = detections

        return result

    # ...def generate_detections_one_image(...)

# ...class PTDetector


#%% Command-line driver

# For testing only... you don't really want to run this module directly.

if __name__ == '__main__':

    pass
    
    #%%
    
    import os
    from megadetector.visualization import visualization_utils as vis_utils
    
    model_file = os.environ['MDV5A']
    im_file = os.path.expanduser('~/git/MegaDetector/images/nacti.jpg')

    detector = PTDetector(model_file)
    image = vis_utils.load_image(im_file)

    res = detector.generate_detections_one_image(image, im_file, detection_threshold=0.00001)
    print(res)



"""

remove_exif.py

Removes all EXIF/IPTC/XMP metadata from a folder of images, without making 
backup copies, using pyexiv2.  Ignores non-jpeg images.

This module is rarely used, and pyexiv2 is not thread-safe, so pyexiv2 is not
included in package-level dependency lists.  YMMV.

"""

#%% Imports and constants

import os
import glob

from multiprocessing.pool import Pool as Pool
from tqdm import tqdm


#%% Support functions

# Pyexif2 is not thread safe, do not call this function in parallel within a process
#
# Parallelizing across processes is fine.
def remove_exif_from_image(fn):

    import pyexiv2
    
    try:
        img = pyexiv2.Image(fn)
        img.clear_exif()
        img.clear_iptc()
        img.clear_xmp()
        img.close()        
    except Exception as e:
        print('EXIF error on {}: {}'.format(fn,str(e)))
    
    return True


#%% Remove EXIF data

def remove_exif(image_base_folder,recursive=True,n_processes=1):
    """
    Removes all EXIF/IPTC/XMP metadata from a folder of images, without making 
    backup copies, using pyexiv2.  Ignores non-jpeg images.
    
    Args:
        image_base_folder (str): the folder from which we should remove EXIF data
        recursive (bool, optional): whether to process [image_base_folder] recursively
        n_processes (int, optional): number of concurrent workers.  Because pyexiv2 is not
            thread-safe, only process-based parallelism is supported.        
    """
    try:
        import pyexiv2 #noqa
    except:
        print('pyexiv2 not available; try "pip install pyexiv2"')
        raise

        
    ##%% List files

    assert os.path.isdir(image_base_folder), \
        'Could not find folder {}'.format(image_base_folder)
    all_files = [f for f in glob.glob(image_base_folder+ "*/**", recursive=recursive)]
    image_files = [s for s in all_files if \
                   (s.lower().endswith('.jpg') or s.lower().endswith('.jpeg'))]
        

    ##%% Remove EXIF data (execution)

    if n_processes == 1:
        
        # fn = image_files[0]
        for fn in tqdm(image_files):
            remove_exif_from_image(fn)
            
    else:
        # pyexiv2 is not thread-safe, so we need to use processes
        print('Starting parallel process pool with {} workers'.format(n_processes))
        pool = Pool(n_processes)
        _ = list(tqdm(pool.imap(remove_exif_from_image,image_files),total=len(image_files)))
            
# ...remove_exif(...)


#%% Command-line driver

## TODO
